@@ -5,20 +5,18 @@ import struct
 import sys
 
 from peewee import *
-from peewee import Alias
+from peewee import ColumnBase
 from peewee import EnclosedNodeList
 from peewee import Entity
 from peewee import Expression
-from peewee import Function
+from peewee import Insert
 from peewee import Node
 from peewee import NodeList
 from peewee import OP
-from peewee import Select
 from peewee import VirtualField
 from peewee import merge_dict
-from peewee import sort_models
 from peewee import sqlite3
-from peewee import text_type
+from peewee import __sqlite_version__
 try:
     from playhouse._sqlite_ext import (
         backup,
@@ -30,6 +28,7 @@ try:
         register_rank_functions,
         sqlite_get_db_status,
         sqlite_get_status,
+        TableFunction,
         ZeroBlob,
     )
     CYTHON_SQLITE_EXTENSIONS = True
@@ -72,117 +71,111 @@ class AutoIncrementField(AutoField):
         return NodeList((node_list, SQL('AUTOINCREMENT')))
 
 
-class JSONPath(Function):
-    def __init__(self, field, path):
-        self.field = field
-        self._path = path
-        self.name = 'json_extract'
-        self.arguments = (self.field, self.sql_path)
+class TDecimalField(DecimalField):
+    field_type = 'TEXT'
+    def get_modifiers(self): pass
+
+
+class JSONPath(ColumnBase):
+    def __init__(self, field, path=None):
+        super(JSONPath, self).__init__()
+        self._field = field
+        self._path = path or ()
 
     @property
     def path(self):
-        return ''.join(self._path)
-
-    @property
-    def sql_path(self):
-        path = self.path
-        if path[0] == '[':
-            path = '$%s' % path
-        else:
-            path = '$.%s' % path
-
-    def update_path(self, part):
-        new_path = list(self._path)
-        if isinstance(part, int):
-            new_path.append('[%s]' % part)
-        else:
-            new_path.append('.%s' % part)
-        return new_path
+        return Value('$%s' % ''.join(self._path))
 
     def __getitem__(self, idx):
-        return JSONPath(self.field, self.update_path(idx))
+        if isinstance(idx, int):
+            item = '[%s]' % idx
+        else:
+            item = '.%s' % idx
+        return JSONPath(self._field, self._path + (item,))
+
+    def set(self, value, as_json=None):
+        if as_json or isinstance(value, (list, dict)):
+            value = fn.json(self._field._json_dumps(value))
+        return fn.json_set(self._field, self.path, value)
+
+    def update(self, value):
+        return self.set(fn.json_patch(self, self._field._json_dumps(value)))
+
+    def remove(self):
+        return fn.json_remove(self._field, self.path)
+
+    def json_type(self):
+        return fn.json_type(self._field, self.path)
+
+    def length(self):
+        return fn.json_array_length(self._field, self.path)
+
+    def children(self):
+        return fn.json_each(self._field, self.path)
+
+    def tree(self):
+        return fn.json_tree(self._field, self.path)
+
+    def __sql__(self, ctx):
+        return ctx.sql(fn.json_extract(self._field, self.path)
+                       if self._path else self._field)
 
 
 class JSONField(TextField):
+    field_type = 'JSON'
+    unpack = False
+
+    def __init__(self, json_dumps=None, json_loads=None, **kwargs):
+        self._json_dumps = json_dumps or json.dumps
+        self._json_loads = json_loads or json.loads
+        super(JSONField, self).__init__(**kwargs)
+
     def python_value(self, value):
         if value is not None:
             try:
-                return json.loads(value)
+                return self._json_loads(value)
             except (TypeError, ValueError):
                 return value
 
     def db_value(self, value):
         if value is not None:
-            return json.dumps(value)
+            if not isinstance(value, Node):
+                value = fn.json(self._json_dumps(value))
+            return value
 
-    def clean_path(self, path):
-        if path.startswith('[') or not path:
-            return '$%s' % path
-        return '$.%s' % path
+    def _e(op):
+        def inner(self, rhs):
+            if isinstance(rhs, (list, dict)):
+                rhs = Value(rhs, converter=self.db_value, unpack=False)
+            return Expression(self, op, rhs)
+        return inner
+    __eq__ = _e(OP.EQ)
+    __ne__ = _e(OP.NE)
+    __gt__ = _e(OP.GT)
+    __ge__ = _e(OP.GTE)
+    __lt__ = _e(OP.LT)
+    __le__ = _e(OP.LTE)
+    __hash__ = Field.__hash__
 
-    def clean_paths(self, paths):
-        return [self.clean_path(path) for path in paths]
+    def __getitem__(self, item):
+        return JSONPath(self)[item]
 
-    def length(self, *paths):
-        if paths:
-            return fn.json_array_length(self, *self.clean_paths(paths))
-        return fn.json_array_length(self)
-
-    def extract(self, *paths):
-        return fn.json_extract(self, *self.clean_paths(paths))
-
-    def __getitem__(self, path):
-        if not isinstance(paths, tuple):
-            paths = (paths,)
-        return self.extract(*paths)
-
-    def _value_for_insertion(self, value):
-        if isinstance(value, (list, tuple, dict)):
-            return fn.json(json.dumps(value))
-        return value
-
-    def _insert_like(self, fn, pairs, mapping):
-        npairs = len(pairs)
-        if npairs == 1 and isinstance(pairs[0], dict):
-            # We were passed a single dictionary to update the field with.
-            mapping.update(pairs[0])
-            npairs = 0
-        elif npairs % 2 != 0:
-            raise ValueError('Unequal key and value parameters.')
-
-        accum = []
-        if npairs:
-            for i in range(0, npairs, 2):
-                accum.append(self.clean_path(pairs[i]))
-                accum.append(self._value_for_insertion(pairs[i + 1]))
-        if mapping:
-            for key in mapping:
-                val = mapping[key]
-                accum.append(self.clean_path(key))
-                accum.append(self._value_for_insertion(val))
-        return fn(self, *accum)
-
-    def insert(self, *pairs, **data):
-        return self._insert_like(fn.json_insert, pairs, data)
-
-    def replace(self, *pairs, **data):
-        return self._insert_like(fn.json_replace, pairs, data)
-
-    def set(self, *pairs, **data):
-        return self._insert_like(fn.json_set, pairs, data)
-
-    def remove(self, *paths):
-        return fn.json_remove(self, *self.clean_paths(paths))
+    def set(self, value, as_json=None):
+        return JSONPath(self).set(value, as_json)
 
     def update(self, data):
-        return fn.json_patch(self, self._value_for_insertion(data))
+        return JSONPath(self).update(data)
 
-    def json_type(self, path=None):
-        if path:
-            return fn.json_type(self, self.clean_path(path))
+    def remove(self):
+        return JSONPath(self).remove()
+
+    def json_type(self):
         return fn.json_type(self)
 
-    def children(self, path=None):
+    def length(self):
+        return fn.json_array_length(self)
+
+    def children(self):
         """
         Schema of `json_each` and `json_tree`:
 
@@ -197,13 +190,9 @@ class JSONField(TextField):
         json JSON hidden (1st input parameter to function)
         root TEXT hidden (2nd input parameter, path at which to start)
         """
-        if path:
-            return fn.json_each(self, self.clean_path(path))
         return fn.json_each(self)
 
-    def tree(self, path=None):
-        if path:
-            return fn.json_tree(self, self.clean_path(path))
+    def tree(self):
         return fn.json_tree(self)
 
 
@@ -215,12 +204,38 @@ class SearchField(Field):
         super(SearchField, self).__init__(unindexed=unindexed,
                                           column_name=column_name, null=True)
 
+    def match(self, term):
+        return match(self, term)
+
+    @property
+    def fts_column_index(self):
+        if not hasattr(self, '_fts_column_index'):
+            search_fields = [f.name for f in self.model._meta.sorted_fields
+                             if isinstance(f, SearchField)]
+            self._fts_column_index = search_fields.index(self.name)
+        return self._fts_column_index
+
+    def highlight(self, left, right):
+        column_idx = self.fts_column_index
+        return fn.highlight(self.model._meta.entity, column_idx, left, right)
+
+    def snippet(self, left, right, over_length='...', max_tokens=16):
+        if not (0 < max_tokens < 65):
+            raise ValueError('max_tokens must be between 1 and 64 (inclusive)')
+        column_idx = self.fts_column_index
+        return fn.snippet(self.model._meta.entity, column_idx, left, right,
+                          over_length, max_tokens)
+
 
 class VirtualTableSchemaManager(SchemaManager):
     def _create_virtual_table(self, safe=True, **options):
         options = self.model.clean_options(
             merge_dict(self.model._meta.options, options))
 
+        # Structure:
+        # CREATE VIRTUAL TABLE <model>
+        # USING <extension_module>
+        # ([prefix_arguments, ...] fields, ... [arguments, ...], [options...])
         ctx = self._create_context()
         ctx.literal('CREATE VIRTUAL TABLE ')
         if safe:
@@ -454,7 +469,7 @@ _alphanum = (set('\t ,"(){}*:_+0123456789') |
              set(_alphabet.upper()) |
              set((chr(26),)))
 _invalid_ascii = set(chr(p) for p in range(128) if chr(p) not in _alphanum)
-_quote_re = re.compile('(?:[^\s"]|"(?:\\.|[^"])*")+')
+_quote_re = re.compile(r'(?:[^\s"]|"(?:\\.|[^"])*")+')
 
 
 class FTS5Model(BaseFTSModel):
@@ -606,10 +621,7 @@ class FTS5Model(BaseFTSModel):
 
     @classmethod
     def rank(cls, *args):
-        if args:
-            return cls.bm25(*args)
-        else:
-            return SQL('rank')
+        return cls.bm25(*args) if args else SQL('rank')
 
     @classmethod
     def bm25(cls, *weights):
@@ -655,7 +667,7 @@ class FTS5Model(BaseFTSModel):
                 .order_by(order_by))
 
     @classmethod
-    def _fts_cmd(cls, cmd, **extra_params):
+    def _fts_cmd_sql(cls, cmd, **extra_params):
         tbl = cls._meta.entity
         columns = [tbl]
         values = [cmd]
@@ -663,14 +675,17 @@ class FTS5Model(BaseFTSModel):
             columns.append(Entity(key))
             values.append(value)
 
-        inner_clause = EnclosedClause(tbl)
-        clause = Clause(
+        return NodeList((
             SQL('INSERT INTO'),
             cls._meta.entity,
-            EnclosedClause(*columns),
+            EnclosedNodeList(columns),
             SQL('VALUES'),
-            EnclosedClause(*values))
-        return cls._meta.database.execute(clause)
+            EnclosedNodeList(values)))
+
+    @classmethod
+    def _fts_cmd(cls, cmd, **extra_params):
+        query = cls._fts_cmd_sql(cmd, **extra_params)
+        return cls._meta.database.execute(query)
 
     @classmethod
     def automerge(cls, level):
@@ -893,10 +908,7 @@ class LSMTable(VirtualModel):
             pk)
 
         if is_single:
-            try:
-                row = query.get()
-            except cls.DoesNotExist:
-                raise KeyError(pk)
+            row = query.get()
             return row[1] if cls._meta._value_field is not None else row
         else:
             return query
@@ -928,13 +940,15 @@ class LSMTable(VirtualModel):
 OP.MATCH = 'MATCH'
 
 def _sqlite_regexp(regex, value):
-    return re.search(regex, value, re.I) is not None
+    return re.search(regex, value) is not None
 
 
 class SqliteExtDatabase(SqliteDatabase):
+    returning_clause = __sqlite_version__ >= (3, 35, 0)
+
     def __init__(self, database, c_extensions=None, rank_functions=True,
                  hash_functions=False, regexp_function=False,
-                 bloomfilter=False, *args, **kwargs):
+                 bloomfilter=False, json_contains=False, *args, **kwargs):
         super(SqliteExtDatabase, self).__init__(database, *args, **kwargs)
         self._row_factory = None
 
@@ -949,6 +963,8 @@ class SqliteExtDatabase(SqliteDatabase):
             else:
                 self.register_function(bm25, 'fts_bm25')
                 self.register_function(rank, 'fts_rank')
+                self.register_function(bm25, 'fts_bm25f')  # Fall back to bm25.
+                self.register_function(bm25, 'fts_lucene')
         if hash_functions:
             if not prefer_c:
                 raise ValueError('C extension required to register hash '
@@ -957,7 +973,11 @@ class SqliteExtDatabase(SqliteDatabase):
         if regexp_function:
             self.register_function(_sqlite_regexp, 'regexp', 2)
         if bloomfilter:
+            if not prefer_c:
+                raise ValueError('C extension required to use bloomfilter.')
             register_bloomfilter(self)
+        if json_contains:
+            self.register_function(_json_contains, 'json_contains')
 
         self._c_extensions = prefer_c
 
@@ -966,6 +986,16 @@ class SqliteExtDatabase(SqliteDatabase):
         if self._row_factory:
             conn.row_factory = self._row_factory
 
+    def last_insert_id(self, cursor, query_type=None):
+        if not self.returning_clause:
+            return cursor.lastrowid
+        elif query_type == Insert.SIMPLE:
+            try:
+                return cursor[0][0]
+            except (AttributeError, IndexError, TypeError):
+                return cursor.lastrowid
+        return cursor
+
     def row_factory(self, fn):
         self._row_factory = fn
 
@@ -996,8 +1026,8 @@ if CYTHON_SQLITE_EXTENSIONS:
 
     def __status__(flag, return_highwater=False):
         """
-        Expose a sqlite3_status() call for a particular flag as a property of the
-        Database object.
+        Expose a sqlite3_status() call for a particular flag as a property of
+        the Database object.
         """
         def getter(self):
             result = sqlite_get_status(flag)
@@ -1006,11 +1036,13 @@ if CYTHON_SQLITE_EXTENSIONS:
 
     def __dbstatus__(flag, return_highwater=False, return_current=False):
         """
-        Expose a sqlite3_dbstatus() call for a particular flag as a property of the
-        Database instance. Unlike sqlite3_status(), the dbstatus properties pertain
-        to the current connection.
+        Expose a sqlite3_dbstatus() call for a particular flag as a property of
+        the Database instance. Unlike sqlite3_status(), the dbstatus properties
+        pertain to the current connection.
         """
         def getter(self):
+            if self._state.conn is None:
+                raise ImproperlyConfigured('database connection not opened.')
             result = sqlite_get_db_status(self._state.conn, flag)
             if return_current:
                 return result[0]
@@ -1047,7 +1079,8 @@ if CYTHON_SQLITE_EXTENSIONS:
             if self._update_hook is not None:
                 self._conn_helper.set_update_hook(self._update_hook)
             if self._replace_busy_handler:
-                self._conn_helper.set_busy_handler(self.timeout or 5000)
+                timeout = self._timeout or 5
+                self._conn_helper.set_busy_handler(timeout * 1000)
 
         def on_commit(self, fn):
             self._commit_hook = fn
@@ -1078,11 +1111,14 @@ if CYTHON_SQLITE_EXTENSIONS:
         def autocommit(self):
             return self._conn_helper.autocommit()
 
-        def backup(self, destination):
-            return backup(self.connection(), destination.connection())
+        def backup(self, destination, pages=None, name=None, progress=None):
+            return backup(self.connection(), destination.connection(),
+                          pages=pages, name=name, progress=progress)
 
-        def backup_to_file(self, filename):
-            return backup_to_file(self.connection(), filename)
+        def backup_to_file(self, filename, pages=None, name=None,
+                           progress=None):
+            return backup_to_file(self.connection(), filename, pages=pages,
+                                  name=name, progress=progress)
 
         def blob_open(self, table, column, rowid, read_only=False):
             return Blob(self, table, column, rowid, read_only)
@@ -1123,6 +1159,15 @@ def _parse_match_info(buf):
     bufsize = len(buf)  # Length in bytes.
     return [struct.unpack('@I', buf[i:i+4])[0] for i in range(0, bufsize, 4)]
 
+def get_weights(ncol, raw_weights):
+    if not raw_weights:
+        return [1] * ncol
+    else:
+        weights = [0] * ncol
+        for i, weight in enumerate(raw_weights):
+            weights[i] = weight
+    return weights
+
 # Ranking implementation, which parse matchinfo.
 def rank(raw_match_info, *raw_weights):
     # Handle match_info called w/default args 'pcx' - based on the example rank
@@ -1131,13 +1176,14 @@ def rank(raw_match_info, *raw_weights):
     score = 0.0
 
     p, c = match_info[:2]
-    if not raw_weights:
-        weights = [1] * c
-    else:
-        weights = [0] * c
-        for i, weight in enumerate(raw_weights):
-            weights[i] = weight
+    weights = get_weights(c, raw_weights)
 
+    # matchinfo X value corresponds to, for each phrase in the search query, a
+    # list of 3 values for each column in the search table.
+    # So if we have a two-phrase search query and three columns of data, the
+    # following would be the layout:
+    # p0 : c0=[0, 1, 2],   c1=[3, 4, 5],    c2=[6, 7, 8]
+    # p1 : c0=[9, 10, 11], c1=[12, 13, 14], c2=[15, 16, 17]
     for phrase_num in range(p):
         phrase_info_idx = 2 + (phrase_num * c * 3)
         for col_num in range(c):
@@ -1146,9 +1192,15 @@ def rank(raw_match_info, *raw_weights):
                 continue
 
             col_idx = phrase_info_idx + (col_num * 3)
-            x1, x2 = match_info[col_idx:col_idx + 2]
-            if x1 > 0:
-                score += weight * (float(x1) / x2)
+
+            # The idea is that we count the number of times the phrase appears
+            # in this column of the current row, compared to how many times it
+            # appears in this column across all rows. The ratio of these values
+            # provides a rough way to score based on "high value" terms.
+            row_hits = match_info[col_idx]
+            all_rows_hits = match_info[col_idx + 1]
+            if row_hits > 0:
+                score += weight * (float(row_hits) / all_rows_hits)
 
     return -score
 
@@ -1167,19 +1219,41 @@ def bm25(raw_match_info, *args):
     B = 0.75
     score = 0.0
 
-    P_O, C_O, N_O, A_O = range(4)
-    term_count = match_info[P_O]
+    P_O, C_O, N_O, A_O = range(4)  # Offsets into the matchinfo buffer.
+    term_count = match_info[P_O]  # n
     col_count = match_info[C_O]
-    total_docs = match_info[N_O]
+    total_docs = match_info[N_O]  # N
     L_O = A_O + col_count
     X_O = L_O + col_count
 
-    if not args:
-        weights = [1] * col_count
-    else:
-        weights = [0] * col_count
-        for i, weight in enumerate(args):
-            weights[i] = args[i]
+    # Worked example of pcnalx for two columns and two phrases, 100 docs total.
+    # {
+    #   p  = 2
+    #   c  = 2
+    #   n  = 100
+    #   a0 = 4   -- avg number of tokens for col0, e.g. title
+    #   a1 = 40  -- avg number of tokens for col1, e.g. body
+    #   l0 = 5   -- curr doc has 5 tokens in col0
+    #   l1 = 30  -- curr doc has 30 tokens in col1
+    #
+    #   x000     -- hits this row for phrase0, col0
+    #   x001     -- hits all rows for phrase0, col0
+    #   x002     -- rows with phrase0 in col0 at least once
+    #
+    #   x010     -- hits this row for phrase0, col1
+    #   x011     -- hits all rows for phrase0, col1
+    #   x012     -- rows with phrase0 in col1 at least once
+    #
+    #   x100     -- hits this row for phrase1, col0
+    #   x101     -- hits all rows for phrase1, col0
+    #   x102     -- rows with phrase1 in col0 at least once
+    #
+    #   x110     -- hits this row for phrase1, col1
+    #   x111     -- hits all rows for phrase1, col1
+    #   x112     -- rows with phrase1 in col1 at least once
+    # }
+
+    weights = get_weights(col_count, args)
 
     for i in range(term_count):
         for j in range(col_count):
@@ -1187,28 +1261,64 @@ def bm25(raw_match_info, *args):
             if weight == 0:
                 continue
 
-            avg_length = float(match_info[A_O + j])
-            doc_length = float(match_info[L_O + j])
-            if avg_length == 0:
-                D = 0
-            else:
-                D = 1 - B + (B * (doc_length / avg_length))
+            x = X_O + (3 * (j + i * col_count))
+            term_frequency = float(match_info[x])  # f(qi, D)
+            docs_with_term = float(match_info[x + 2])  # n(qi)
 
-            x = X_O + (3 * j * (i + 1))
-            term_frequency = float(match_info[x])
-            docs_with_term = float(match_info[x + 2])
-
-            idf = max(
-                math.log(
+            # log( (N - n(qi) + 0.5) / (n(qi) + 0.5) )
+            idf = math.log(
                     (total_docs - docs_with_term + 0.5) /
-                    (docs_with_term + 0.5)),
-                0)
-            denom = term_frequency + (K * D)
-            if denom == 0:
-                rhs = 0
-            else:
-                rhs = (term_frequency * (K + 1)) / denom
+                    (docs_with_term + 0.5))
+            if idf <= 0.0:
+                idf = 1e-6
 
-            score += (idf * rhs) * weight
+            doc_length = float(match_info[L_O + j])  # |D|
+            avg_length = float(match_info[A_O + j]) or 1.  # avgdl
+            ratio = doc_length / avg_length
+
+            num = term_frequency * (K + 1.0)
+            b_part = 1.0 - B + (B * ratio)
+            denom = term_frequency + (K * b_part)
+
+            pc_score = idf * (num / denom)
+            score += (pc_score * weight)
 
     return -score
+
+
+def _json_contains(src_json, obj_json):
+    stack = []
+    try:
+        stack.append((json.loads(obj_json), json.loads(src_json)))
+    except:
+        # Invalid JSON!
+        return False
+
+    while stack:
+        obj, src = stack.pop()
+        if isinstance(src, dict):
+            if isinstance(obj, dict):
+                for key in obj:
+                    if key not in src:
+                        return False
+                    stack.append((obj[key], src[key]))
+            elif isinstance(obj, list):
+                for item in obj:
+                    if item not in src:
+                        return False
+            elif obj not in src:
+                return False
+        elif isinstance(src, list):
+            if isinstance(obj, dict):
+                return False
+            elif isinstance(obj, list):
+                try:
+                    for i in range(len(obj)):
+                        stack.append((obj[i], src[i]))
+                except IndexError:
+                    return False
+            elif obj not in src:
+                return False
+        elif obj != src:
+            return False
+    return True
